# Copyright (c) 2009-2012 VMware, Inc.

require 'spec_helper'

describe Bosh::AwsCloud::Cloud do

  let(:zones) { [double('us-east-1a', :name => 'us-east-1a')] }
  let(:volume) { double('volume', :id => 'v-foobar') }
  let(:instance) { double('instance', id: 'i-test', availability_zone: 'foobar-land') }

  before do
    allow(Bosh::AwsCloud::ResourceWait).to receive(:for_volume).with(volume: volume, state: :available)
    @cloud = mock_cloud do |_ec2|
      @ec2 = _ec2
      allow(@ec2.volumes).to receive(:create) { volume }
      allow(@ec2).to receive_messages(:availability_zones => zones)
      allow(@ec2).to receive_messages(instances: double('instances', :[] => instance))
    end
  end

  it 'creates an EC2 volume' do
    expect(@cloud.create_disk(2048, {})).to eq('v-foobar')
    expect(@ec2.volumes).to have_received(:create) do |params|
      expect(params[:size]).to eq(2)
    end
  end

  it 'rounds up disk size' do
    @cloud.create_disk(2049, {})
    expect(@ec2.volumes).to have_received(:create) do |params|
      expect(params[:size]).to eq(3)
    end
  end

  it 'checks that the disk size is greater than 0' do
    expect {
      @cloud.create_disk(0, {})
    }.to raise_error(Bosh::Clouds::CloudError, /disk size must be greater than 0/)
  end

  it 'puts disk in the same AZ as an instance' do
    @cloud.create_disk(1024, {}, 'i-test')

    expect(@ec2.volumes).to have_received(:create) do |params|
      expect(params[:availability_zone]).to eq('foobar-land')
    end
  end

  it 'should pick a random availability zone when no instance is given' do
    @cloud.create_disk(2048, {})
    expect(@ec2.volumes).to have_received(:create) do |params|
      expect(params[:availability_zone]).to eq('us-east-1a')
    end
  end

  context 'cloud properties' do
    describe 'volume type' do
      it 'defaults to standard' do
        @cloud.create_disk(2048, {})

        expect(@ec2.volumes).to have_received(:create) do |params|
          expect(params[:volume_type]).to eq('standard')
        end
      end

      it 'is pulled from cloud properties' do
        @cloud.create_disk(2048, { 'type' => 'gp2' })

        expect(@ec2.volumes).to have_received(:create) do |params|
          expect(params[:volume_type]).to eq('gp2')
        end
      end
    end

    describe 'encryption' do
      it 'defaults to unencrypted' do
        @cloud.create_disk(2048, {})

        expect(@ec2.volumes).to have_received(:create) do |params|
          expect(params[:encrypted]).to eq(false)
        end
      end

      it 'passes through encryped => true' do
        @cloud.create_disk(2048, { 'encrypted' => true })

        expect(@ec2.volumes).to have_received(:create) do |params|
          expect(params[:encrypted]).to eq(true)
        end
      end
    end
  end
end
